@@ -6,6 +6,8 @@
  *
  */
 
+#define _XOPEN_SOURCE /* needed for strptime */
+
 #include <stdio.h>
 #include <unistd.h>
 #include <stdlib.h>
@@ -80,6 +82,8 @@ FILE *open_pager(void)
 	char buf[80];
 
 	signal(SIGPIPE, SIG_IGN);
+	if (!isatty(1))
+		return stdout;
 	if (!pager)
 		pager = getenv("PAGER");
 	if (!pager)
@@ -119,7 +123,7 @@ ext2_ino_t string_to_inode(char *str)
 
 	retval = ext2fs_namei(current_fs, root, cwd, str, &ino);
 	if (retval) {
-		com_err(str, retval, "");
+		com_err(str, retval, 0);
 		return 0;
 	}
 	return ino;
@@ -188,7 +192,7 @@ char *time_to_string(__u32 cl)
 {
 	static int	do_gmt = -1;
 	time_t		t = (time_t) cl;
-	char *		tz;
+	const char	*tz;
 
 	if (do_gmt == -1) {
 		/* The diet libc doesn't respect the TZ environemnt variable */
@@ -208,7 +212,7 @@ char *time_to_string(__u32 cl)
 extern time_t string_to_time(const char *arg)
 {
 	struct	tm	ts;
-	unsigned long	ret;
+	time_t		ret;
 	char *tmp;
 
 	if (strcmp(arg, "now") == 0) {
@@ -227,14 +231,15 @@ extern time_t string_to_time(const char *arg)
 	    ts.tm_min > 59 || ts.tm_sec > 61)
 		ts.tm_mday = 0;
 #endif
-	if (ts.tm_mday == 0) {
+	ts.tm_isdst = -1;
+	ret = mktime(&ts);
+	if (ts.tm_mday == 0 || ret == ((time_t) -1)) {
 		/* Try it as an integer... */
-
 		ret = strtoul(arg, &tmp, 0);
 		if (*tmp)
 			return ((time_t) -1);
 	}
-	return mktime(&ts);
+	return ret;
 }
 
 /*
@@ -272,10 +277,8 @@ int strtoblk(const char *cmd, const char *str, blk_t *ret)
 
 	blk = parse_ulong(str, cmd, "block number", &err);
 	*ret = blk;
-	if (err == 0 && blk == 0) {
-		com_err(cmd, 0, "Invalid block number 0");
-		err = 1;
-	}
+	if (err)
+		com_err(cmd, 0, "Invalid block number: %s", str);
 	return err;
 }
 
@@ -325,7 +328,7 @@ int common_inode_args_process(int argc, char *argv[],
  * This is a helper function used by do_freeb, do_setb, and do_testb
  */
 int common_block_args_process(int argc, char *argv[],
-			      blk_t *block, int *count)
+			      blk_t *block, blk_t *count)
 {
 	int	err;
 
@@ -335,6 +338,11 @@ int common_block_args_process(int argc, char *argv[],
 
 	if (strtoblk(argv[0], argv[1], block))
 		return 1;
+	if (*block == 0) {
+		com_err(argv[0], 0, "Invalid block number 0");
+		err = 1;
+	}
+
 	if (argc > 2) {
 		*count = parse_ulong(argv[2], argv[0], "count", &err);
 		if (err)
