@@ -126,7 +126,7 @@ static void write_image_file(ext2_filsys fs, int fd)
 	hdr.magic_number = EXT2_ET_MAGIC_E2IMAGE;
 	strcpy(hdr.magic_descriptor, "Ext2 Image 1.0");
 	gethostname(hdr.fs_hostname, sizeof(hdr.fs_hostname));
-	strncat(hdr.fs_device_name, device_name, sizeof(hdr.fs_device_name));
+	strncpy(hdr.fs_device_name, device_name, sizeof(hdr.fs_device_name)-1);
 	hdr.fs_device_name[sizeof(hdr.fs_device_name) - 1] = 0;
 	hdr.fs_blocksize = fs->blocksize;
 	
@@ -330,6 +330,7 @@ static void write_block(int fd, char *buf, int sparse_offset,
 				err = 0;
 			com_err(program_name, err, "error writing block %u",
 				block);
+			exit(1);
 		}
 	}
 }
@@ -346,9 +347,8 @@ static void scramble_dir_block(ext2_filsys fs, blk_t blk, char *buf)
 	for (p = buf; p < end-8; p += rec_len) {
 		dirent = (struct ext2_dir_entry_2 *) p;
 		rec_len = dirent->rec_len;
-#ifdef EXT2FS_ENABLE_SWAPFS
-		if (fs->flags & EXT2_FLAG_SWAP_BYTES) 
-			rec_len = ext2fs_swab16(rec_len);
+#ifdef WORDS_BIGENDIAN
+		rec_len = ext2fs_swab16(rec_len);
 #endif
 #if 0
 		printf("rec_len = %d, name_len = %d\n", rec_len, dirent->name_len);
@@ -359,8 +359,7 @@ static void scramble_dir_block(ext2_filsys fs, blk_t blk, char *buf)
 			       "bad rec_len (%d)\n", (unsigned long) blk, 
 			       rec_len);
 			rec_len = end - p;
-#ifdef EXT2FS_ENABLE_SWAPFS
-			if (fs->flags & EXT2_FLAG_SWAP_BYTES) 
+#ifdef WORDS_BIGENDIAN
 				dirent->rec_len = ext2fs_swab16(rec_len);
 #endif
 			continue;
@@ -515,8 +514,9 @@ static void write_raw_image_file(ext2_filsys fs, int fd, int scramble_flag)
 		    (LINUX_S_ISLNK(inode.i_mode) &&
 		     ext2fs_inode_has_valid_blocks(&inode)) ||
 		    ino == fs->super->s_journal_inum) {
-			retval = ext2fs_block_iterate2(fs, ino, 0, 
-				       block_buf, process_dir_block, &pb);
+			retval = ext2fs_block_iterate2(fs, ino,
+					BLOCK_FLAG_READ_ONLY, block_buf,
+					process_dir_block, &pb);
 			if (retval) {
 				com_err(program_name, retval,
 					"while iterating over inode %u",
@@ -524,11 +524,12 @@ static void write_raw_image_file(ext2_filsys fs, int fd, int scramble_flag)
 				exit(1);
 			}
 		} else {
-			if (inode.i_block[EXT2_IND_BLOCK] ||
+			if ((inode.i_flags & EXT4_EXTENTS_FL) ||
+			    inode.i_block[EXT2_IND_BLOCK] ||
 			    inode.i_block[EXT2_DIND_BLOCK] ||
 			    inode.i_block[EXT2_TIND_BLOCK]) {
 				retval = ext2fs_block_iterate2(fs,
-				       ino, 0, block_buf,
+				       ino, BLOCK_FLAG_READ_ONLY, block_buf,
 				       process_file_block, &pb);
 				if (retval) {
 					com_err(program_name, retval,
@@ -578,8 +579,11 @@ static void install_image(char *device, char *image_fn, int raw_flag)
 		exit(1);
 	}
 
-
+#ifdef HAVE_OPEN64
+	fd = open64(image_fn, O_RDONLY);
+#else
 	fd = open(image_fn, O_RDONLY);
+#endif
 	if (fd < 0) {
 		perror(image_fn);
 		exit(1);
