@@ -81,6 +81,8 @@ static const char *really_wanted[] = {
 	"minix",
 	"ext2",
 	"ext3",
+	"ext4",
+	"ext4dev",
 	"jfs",
 	"reiserfs",
 	"xiafs",
@@ -102,7 +104,7 @@ int doall = 0;
 int noexecute = 0;
 int serialize = 0;
 int skip_root = 0;
-int like_mount = 0;
+int ignore_mounted = 0;
 int notitle = 0;
 int parallel_root = 0;
 int progress = 0;
@@ -364,7 +366,7 @@ static void load_fs_info(const char *filename)
 	
 	fclose(f);
 	
-	if (old_fstab) {
+	if (old_fstab && filesys_info) {
 		fputs(_("\007\007\007"
 		"WARNING: Your /etc/fstab does not contain the fsck passno\n"
 		"	field.  I will kludge around things for you, but you\n"
@@ -451,13 +453,21 @@ static int execute(const char *type, const char *device, const char *mntpt,
 	for (i=0; i <num_args; i++)
 		argv[argc++] = string_copy(args[i]);
 
-	if (progress && !progress_active()) {
+	if (progress) {
 		if ((strcmp(type, "ext2") == 0) ||
-		    (strcmp(type, "ext3") == 0)) {
+		    (strcmp(type, "ext3") == 0) ||
+		    (strcmp(type, "ext4") == 0) ||
+		    (strcmp(type, "ext4dev") == 0)) {
 			char tmp[80];
-			snprintf(tmp, 80, "-C%d", progress_fd);
-			argv[argc++] = string_copy(tmp);
-			inst->flags |= FLAG_PROGRESS;
+
+			tmp[0] = 0;
+			if (!progress_active()) {
+				snprintf(tmp, 80, "-C%d", progress_fd);
+				inst->flags |= FLAG_PROGRESS;
+			} else if (progress_fd)
+				snprintf(tmp, 80, "-C%d", progress_fd * -1);
+			if (tmp[0])
+				argv[argc++] = string_copy(tmp);
 		}
 	}
 
@@ -615,13 +625,16 @@ static struct fsck_instance *wait_one(int flags)
 		status = EXIT_ERROR;
 	}
 	inst->exit_status = status;
+	inst->flags |= FLAG_DONE;
 	if (progress && (inst->flags & FLAG_PROGRESS) &&
 	    !progress_active()) {
 		for (inst2 = instance_list; inst2; inst2 = inst2->next) {
 			if (inst2->flags & FLAG_DONE)
 				continue;
 			if (strcmp(inst2->type, "ext2") &&
-			    strcmp(inst2->type, "ext3"))
+			    strcmp(inst2->type, "ext3") &&
+			    strcmp(inst2->type, "ext4") &&
+			    strcmp(inst2->type, "ext4dev"))
 				continue;
 			/*
 			 * If we've just started the fsck, wait a tiny
@@ -800,7 +813,7 @@ static void compile_fs_type(char *fs_type, struct fs_type_compile *cmp)
  * This function returns true if a particular option appears in a
  * comma-delimited options list
  */
-static int opt_in_list(char *opt, char *optlist)
+static int opt_in_list(const char *opt, char *optlist)
 {
 	char	*list, *s;
 
@@ -973,7 +986,8 @@ static int check_all(NOARGS)
 				break;
 		}
 		if (fs) {
-			if (!skip_root && !ignore(fs)) {
+			if (!skip_root && !ignore(fs) &&
+			    !(ignore_mounted && is_mounted(fs->device))) {
 				fsck_device(fs, 1);
 				status |= wait_many(FLAG_WAIT_ALL);
 				if (status > EXIT_NONDESTRUCT)
@@ -1009,6 +1023,10 @@ static int check_all(NOARGS)
 				not_done_yet++;
 				continue;
 			}
+			if (ignore_mounted && is_mounted(fs->device)) {
+				fs->flags |= FLAG_DONE;
+				continue;
+			}
 			/*
 			 * If a filesystem on a particular device has
 			 * already been spawned, then we need to defer
@@ -1058,7 +1076,7 @@ static int check_all(NOARGS)
 
 static void usage(NOARGS)
 {
-	fputs(_("Usage: fsck [-ANPRTV] [ -C [ fd ] ] [-t fstype] [fs-options] [filesys ...]\n"), stderr);
+	fputs(_("Usage: fsck [-AMNPRTV] [ -C [ fd ] ] [-t fstype] [fs-options] [filesys ...]\n"), stderr);
 	exit(EXIT_USAGE);
 }
 
@@ -1181,7 +1199,7 @@ static void PRS(int argc, char *argv[])
 				notitle++;
 				break;
 			case 'M':
-				like_mount++;
+				ignore_mounted++;
 				break;
 			case 'P':
 				parallel_root++;
@@ -1303,6 +1321,8 @@ int main(int argc, char *argv[])
 			if (!fs)
 				continue;
 		}
+		if (ignore_mounted && is_mounted(fs->device))
+			continue;
 		fsck_device(fs, interactive);
 		if (serialize ||
 		    (max_running && (num_running >= max_running))) {
