@@ -71,8 +71,8 @@ void proceed_question(void)
 	fflush(stderr);
 	fputs(_("Proceed anyway? (y,n) "), stdout);
 	buf[0] = 0;
-	fgets(buf, sizeof(buf), stdin);
-	if (strchr(short_yes, buf[0]) == 0)
+	if (!fgets(buf, sizeof(buf), stdin) ||
+	    strchr(short_yes, buf[0]) == 0)
 		exit(1);
 }
 
@@ -152,17 +152,26 @@ void check_mount(const char *device, int force, const char *type)
 			device);
 		return;
 	}
-	if (!(mount_flags & EXT2_MF_MOUNTED))
-		return;
-
-	fprintf(stderr, _("%s is mounted; "), device);
-	if (force) {
-		fputs(_("mke2fs forced anyway.  Hope /etc/mtab is "
-			"incorrect.\n"), stderr);
-	} else {
+	if (mount_flags & EXT2_MF_MOUNTED) {
+		fprintf(stderr, _("%s is mounted; "), device);
+		if (force > 2) {
+			fputs(_("mke2fs forced anyway.  Hope /etc/mtab is "
+				"incorrect.\n"), stderr);
+			return;
+		}
+	abort_mke2fs:
 		fprintf(stderr, _("will not make a %s here!\n"), type);
 		exit(1);
 	}
+	if (mount_flags & EXT2_MF_BUSY) {
+		fprintf(stderr, _("%s is apparently in use by the system; "),
+			device);
+		if (force > 2) {
+			fputs(_("mke2fs forced anyway.\n"), stderr);
+			return;
+		}
+		goto abort_mke2fs;
+	}
 }
 
 void parse_journal_opts(const char *opts)
@@ -225,8 +234,10 @@ void parse_journal_opts(const char *opts)
 			"\tdevice=<journal device>\n\n"
 			"The journal size must be between "
 			"1024 and 102400 filesystem blocks.\n\n"), stderr);
+		free(buf);
 		exit(1);
 	}
+	free(buf);
 }	
 
 /*
@@ -238,40 +249,32 @@ void parse_journal_opts(const char *opts)
  * in the filesystem.  For very small filesystems, it is not reasonable to
  * have a journal that fills more than half of the filesystem.
  */
-int figure_journal_size(int size, ext2_filsys fs)
+unsigned int figure_journal_size(int size, ext2_filsys fs)
 {
-	blk_t j_blocks;
+	int j_blocks;
 
-	if (fs->super->s_blocks_count < 2048) {
+	j_blocks = ext2fs_default_journal_size(fs->super->s_blocks_count);
+	if (j_blocks < 0) {
 		fputs(_("\nFilesystem too small for a journal\n"), stderr);
 		return 0;
 	}
 	
-	if (size >= 0) {
+	if (size > 0) {
 		j_blocks = size * 1024 / (fs->blocksize	/ 1024);
-		if (j_blocks < 1024 || j_blocks > 102400) {
+		if (j_blocks < 1024 || j_blocks > 10240000) {
 			fprintf(stderr, _("\nThe requested journal "
 				"size is %d blocks; it must be\n"
-				"between 1024 and 102400 blocks.  "
+				"between 1024 and 10240000 blocks.  "
 				"Aborting.\n"),
 				j_blocks);
 			exit(1);
 		}
-		if (j_blocks > fs->super->s_free_blocks_count) {
+		if ((unsigned) j_blocks > fs->super->s_free_blocks_count / 2) {
 			fputs(_("\nJournal size too big for filesystem.\n"),
 			      stderr);
 			exit(1);
 		}
-		return j_blocks;
 	}
-
-	if (fs->super->s_blocks_count < 32768)
-		j_blocks = 1024;
-	else if (fs->super->s_blocks_count < 262144)
-		j_blocks = 4096;
-	else
-		j_blocks = 8192;
-
 	return j_blocks;
 }
 
