@@ -36,8 +36,8 @@ char *program_name, *device_name, *io_options;
 
 static void usage (char *prog)
 {
-	fprintf (stderr, _("Usage: %s [-d debug_flags] [-f] [-F] [-p] "
-			   "device [new_size]\n\n"), prog);
+	fprintf (stderr, _("Usage: %s [-d debug_flags] [-f] [-F] [-M] [-P] "
+			   "[-p] device [new_size]\n\n"), prog);
 
 	exit (1);
 }
@@ -152,6 +152,8 @@ int main (int argc, char ** argv)
 	int		flush = 0;
 	int		force = 0;
 	int		io_flags = 0;
+	int		force_min_size = 0;
+	int		print_min_size = 0;
 	int		fd, ret;
 	blk_t		new_size = 0;
 	blk_t		max_size = 0;
@@ -183,7 +185,7 @@ int main (int argc, char ** argv)
 	if (argc && *argv)
 		program_name = *argv;
 
-	while ((c = getopt (argc, argv, "d:fFhpS:")) != EOF) {
+	while ((c = getopt (argc, argv, "d:fFhMPpS:")) != EOF) {
 		switch (c) {
 		case 'h':
 			usage(program_name);
@@ -194,6 +196,12 @@ int main (int argc, char ** argv)
 		case 'F':
 			flush = 1;
 			break;
+		case 'M':
+			force_min_size = 1;
+			break;
+		case 'P':
+			print_min_size = 1;
+			break;
 		case 'd':
 			flags |= atoi(optarg);
 			break;
@@ -298,6 +306,7 @@ int main (int argc, char ** argv)
 		printf (_("Couldn't find valid filesystem superblock.\n"));
 		exit (1);
 	}
+
 	/*
 	 * Check for compatibility with the feature sets.  We need to
 	 * be more stringent than ext2fs_open().
@@ -307,7 +316,32 @@ int main (int argc, char ** argv)
 			"(%s)", device_name);
 		exit(1);
 	}
+
+	/*
+	 * XXXX   The combination of flex_bg and !resize_inode causes
+	 * major problems for resize2fs, since when the group descriptors
+	 * grow in size this can potentially require multiple inode
+	 * tables to be moved aside to make room, and resize2fs chokes
+	 * rather badly in this scenario.  It's a rare combination,
+	 * except when a filesystem is expanded more than a certain
+	 * size, so for now, we'll just prohibit that combination.
+	 * This is something we should fix eventually, though.
+	 */
+	if ((fs->super->s_feature_incompat & EXT4_FEATURE_INCOMPAT_FLEX_BG) &&
+	    !(fs->super->s_feature_compat & EXT2_FEATURE_COMPAT_RESIZE_INODE)) {
+		com_err(program_name, 0, _("%s: The combination of flex_bg "
+					   "and\n\t!resize_inode features "
+					   "is not supported by resize2fs.\n"),
+			device_name);
+		exit(1);
+	}
 	
+	if (print_min_size) {
+		printf(_("Estimated minimum size of the filesystem: %u\n"),
+		       calculate_minimum_resize_size(fs));
+		exit(0);
+	}
+
 	/* Determine the system page size if possible */
 #ifdef HAVE_SYSCONF
 #if (!defined(_SC_PAGESIZE) && defined(_SC_PAGE_SIZE))
@@ -332,14 +366,11 @@ int main (int argc, char ** argv)
 			_("while trying to determine filesystem size"));
 		exit(1);
 	}
-	if (new_size_str) {
+	if (force_min_size)
+		new_size = calculate_minimum_resize_size(fs);
+	else if (new_size_str) {
 		new_size = parse_num_blocks(new_size_str, 
 					    fs->super->s_log_block_size);
-		if (!new_size) {
-			com_err(program_name, 0, _("bad filesystem size - %s"),
-				new_size_str);
-			exit(1);
-		}
 	} else {
 		new_size = max_size;
 		/* Round down to an even multiple of a pagesize */
@@ -413,10 +444,12 @@ int main (int argc, char ** argv)
 
 	if ((st_buf.st_size > new_file_size) &&
 	    (fd > 0)) {
-#ifdef HAVE_FSTAT64
+#ifdef HAVE_FTRUNCATE64
 		ftruncate64(fd, new_file_size);
 #else
-		ftruncate(fd, (off_t) new_file_size);
+		/* Only truncate if new_file_size doesn't overflow off_t */
+		if (((off_t) new_file_size) == new_file_size)
+			ftruncate(fd, (off_t) new_file_size);
 #endif
 	}
 	if (fd > 0)
