@@ -21,14 +21,18 @@ extern char *program_name;
 errcode_t online_resize_fs(ext2_filsys fs, const char *mtpt, 
 			   blk_t *new_size, int flags EXT2FS_ATTR((unused)))
 {
+#ifdef __linux__
 	struct ext2_new_group_input input;
+	struct ext4_new_group_input input64;
 	struct ext2_super_block *sb = fs->super;
 	unsigned long		new_desc_blocks;
 	ext2_filsys 		new_fs;
 	errcode_t 		retval;
+	double			percent;
 	dgrp_t			i;
 	blk_t			size;
-	int			fd, r_frac, overhead;
+	int			fd, overhead;
+	int			use_old_ioctl = 1;
 
 	printf(_("Filesystem at %s is mounted on %s; "
 		 "on-line resizing required\n"), fs->device_name, mtpt);
@@ -79,7 +83,7 @@ errcode_t online_resize_fs(ext2_filsys fs, const char *mtpt,
 		exit(1);
 	}
 
-	r_frac = ext2fs_div_ceil(100 * sb->s_r_blocks_count, sb->s_blocks_count);
+	percent = (sb->s_r_blocks_count * 100.0) / sb->s_blocks_count;
 
 	retval = ext2fs_read_bitmaps(fs);
 	if (retval)
@@ -126,8 +130,8 @@ errcode_t online_resize_fs(ext2_filsys fs, const char *mtpt,
 				sb->s_first_data_block - 
 				(i * sb->s_blocks_per_group);
 		}
-		input.reserved_blocks = e2p_percent(r_frac, 
-						    input.blocks_count);
+		input.reserved_blocks = (blk_t) (percent * input.blocks_count
+						 / 100.0);
 
 #if 0
 		printf("new block bitmap is at 0x%04x\n", input.block_bitmap);
@@ -146,7 +150,21 @@ errcode_t online_resize_fs(ext2_filsys fs, const char *mtpt,
 		printf("Adding group #%d\n", input.group);
 #endif
 
-		if (ioctl(fd, EXT2_IOC_GROUP_ADD, &input) < 0) {
+		if (use_old_ioctl &&
+		    ioctl(fd, EXT2_IOC_GROUP_ADD, &input) == 0)
+			continue;
+		else
+			use_old_ioctl = 1;
+
+		input64.group = input.group;
+		input64.block_bitmap = input.block_bitmap;
+		input64.inode_bitmap = input.inode_bitmap;
+		input64.inode_table = input.inode_table;
+		input64.blocks_count = input.blocks_count;
+		input64.reserved_blocks = input.reserved_blocks;
+		input64.unused = input.unused;
+
+		if (ioctl(fd, EXT4_IOC_GROUP_ADD, &input64) < 0) {
 			com_err(program_name, errno, 
 				_("While trying to add group #%d"), 
 				input.group);
@@ -158,4 +176,9 @@ errcode_t online_resize_fs(ext2_filsys fs, const char *mtpt,
 	close(fd);
 
 	return 0;
+#else
+	printf(_("Filesystem at %s is mounted on %s, and on-line resizing is"
+		 "not supported on this system.\n"), fs->device_name, mtpt);
+	exit(1);
+#endif
 }
